from typing import Tuple, Dict

from cirro.cli.interactive.utils import ask_yes_no


def gather_auth_config() -> Tuple[str, Dict]:
    auth_method_config = {
        'enable_cache': ask_yes_no("Would you like to save your login? (do not use this on shared devices)")
    }

    return 'ClientAuth', auth_method_config



import logging
import threading
from datetime import datetime, timedelta

import boto3
from pycognito import AWSSRP
from requests.auth import AuthBase

from cirro.api.auth.base import AuthInfo, RequestAuthWrapper

logger = logging.getLogger()


class UsernameAndPasswordAuth(AuthInfo):
    """
    Uses your username & password to authenticate
    Note: this does not work with federated identities (Fred Hutch login)
    You must contact HDC to manually create an account
    """
    def __init__(self, username: str, password: str, region: str, user_pool_id: str, client_id: str):
        self.region = region
        self.client_id = client_id
        self.user_pool_id = user_pool_id
        self._username = username
        self._password = password
        self._auth_result = None
        self._get_token_lock = threading.Lock()
        self.token_expiry = None

    def get_request_auth(self) -> AuthBase:
        return RequestAuthWrapper(lambda: self._get_token()['AccessToken'])

    def get_current_user(self) -> str:
        return self._username

    def _get_token(self):
        if self.token_expiry and self.token_expiry > datetime.now():
            return self._auth_result

        with self._get_token_lock:
            logger.debug('Fetching new token from cognito')
            cognito = boto3.client('cognito-idp', region_name=self.region)
            aws = AWSSRP(username=self._username,
                         password=self._password,
                         pool_id=self.user_pool_id,
                         client_id=self.client_id,
                         client=cognito)
            resp = aws.authenticate_user()
            self._auth_result = resp['AuthenticationResult']
            self.access_token_expiry = datetime.now() + timedelta(seconds=self._auth_result['ExpiresIn'])
            return self._auth_result



from typing import TypedDict, Optional


class Creds(TypedDict):
    AccessKeyId: str
    SecretAccessKey: str
    SessionToken: Optional[str]
    Expiration: Optional[str]


class DeviceTokenResponse(TypedDict):
    device_code: str
    user_code: str
    verification_uri: str
    message: str
    interval: int
    expires_in: int
    expiry: str


class OAuthTokenResponse(TypedDict):
    access_token: str
    refresh_token: str
    id_token: str
    token_type: str
    expires_in: int
    refresh_expires_in: int
    client_id: str
    message: Optional[str]



from typing import NamedTuple, List, Optional, Dict, TypedDict

from cirro.api.models.process import ProcessRecord


class ProcessConfig(TypedDict):
    dynamo: ProcessRecord
    form: Dict
    input: Dict[str, str]
    output: Dict


class WorkflowRepository(NamedTuple):
    display_name: str
    org: str
    repo_name: str
    version: str
    entrypoint: str
    private: bool
    documentation_url: str
    tarball: Optional[str]

    @property
    def repo_path(self):
        return f'{self.org}/{self.repo_name}'


class Column(NamedTuple):
    header: str
    display_name: str
    description: str


class OptimizedOutput(NamedTuple):
    source_pattern: str
    seperator: str
    name: str
    description: str
    documentation_url: str
    columns: List[Column]



from cirro.api.models.reference import Reference
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset


class DataPortalReference(DataPortalAsset):
    """
    Reference data is organized by project, categorized by type.
    """
    name = None

    def __init__(self, ref: Reference):
        self.name = ref.name
        self._access_context = ref.access_context
        self.relative_path = ref.relative_path
        self.absolute_path = f'{self._access_context.domain}/{self.relative_path.strip("/")}'

    def __str__(self):
        return self.name


class DataPortalReferences(DataPortalAssets[DataPortalReference]):
    """Collection of DataPortalReference objects."""
    asset_name = "reference"

    def get_by_id(self, _id: str) -> DataPortalReference:
        raise NotImplementedError("Filtering by ID is not supported, use get_by_name")



import json
import logging
import sys
import threading
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

import boto3
import jwt
import requests
from botocore.exceptions import ClientError
from msal_extensions import FilePersistence
from msal_extensions.persistence import BasePersistence
from requests.auth import AuthBase

from cirro.api.auth.base import AuthInfo, RequestAuthWrapper
from cirro.api.config import Constants
from cirro.api.models.auth import DeviceTokenResponse, OAuthTokenResponse

logger = logging.getLogger()
TOKEN_PATH = Path(Constants.home, '.token.dat').expanduser()


def _build_token_persistence(location, fallback_to_plaintext=False):
    try:
        if sys.platform.startswith('win'):
            from msal_extensions import FilePersistenceWithDataProtection
            return FilePersistenceWithDataProtection(location)
        if sys.platform.startswith('darwin'):
            from msal_extensions import KeychainPersistence
            return KeychainPersistence(location, service_name='cirro-client')
        if sys.platform.startswith('linux'):
            from msal_extensions import LibsecretPersistence
            return LibsecretPersistence(location)
        raise RuntimeError(f"Unsupported platform: {sys.platform}")
    except Exception:
        if not fallback_to_plaintext:
            raise
        logger.debug("Encryption unavailable. Opting in to plain text.")
        return FilePersistence(location)


def _authenticate(client_id: str, auth_endpoint: str):
    params = {'client_id': client_id}
    resp = requests.post(f'{auth_endpoint}/device-code', params=params)
    resp.raise_for_status()
    flow: DeviceTokenResponse = resp.json()
    print(flow['message'])
    device_expiry = datetime.fromisoformat(flow['expiry'])

    params = {
        'client_id': client_id,
        'device_code': flow['device_code'],
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'
    }

    auth_status = 'authorization_pending'
    while auth_status == 'authorization_pending':
        time.sleep(flow['interval'])
        if device_expiry < datetime.now().astimezone():
            raise RuntimeError('Authentication timed out')

        resp = requests.post(f'{auth_endpoint}/token', params=params)
        token_result: OAuthTokenResponse = resp.json()
        auth_status = token_result.get('message')
        logger.debug(auth_status)

        if 'access_token' in token_result:
            return token_result

    raise RuntimeError(f'error authenticating {auth_status}')


class ClientAuth(AuthInfo):
    """
    Authenticates to Cirro by asking
     the user to enter a verification code on the portal website

    :param enable_cache: Optionally enable cache to avoid re-authentication

    Implements the OAuth device code flow
    This is the preferred way to authenticate
    """
    def __init__(self, client_id: str, region: str, auth_endpoint: str, enable_cache=True):
        self.client_id = client_id
        self.region = region
        self._token_info: Optional[OAuthTokenResponse] = None
        self._persistence: Optional[BasePersistence] = None

        if enable_cache:
            self._persistence = _build_token_persistence(str(TOKEN_PATH), fallback_to_plaintext=True)
            self._token_info = self._load_token_info()

        # Check saved token for change in endpoint
        if self._token_info and self._token_info.get('client_id') != client_id:
            logger.debug('Different client ID found, clearing saved token info')
            self._clear_token_info()

        # Check saved token for refresh token expiry
        if self._token_info and self._token_info.get('refresh_expires_in'):
            refresh_expiry_threshold = datetime.fromtimestamp(self._token_info.get('refresh_expires_in'))\
                                       - timedelta(hours=12)
            if refresh_expiry_threshold < datetime.now():
                logger.debug('Refresh token expiry is too soon, re-authenticating')
                self._clear_token_info()

        if not self._token_info:
            self._token_info = _authenticate(client_id=client_id, auth_endpoint=auth_endpoint)

        self._save_token_info()
        self._update_token_metadata()
        self._get_token_lock = threading.Lock()

    def get_request_auth(self) -> AuthBase:
        return RequestAuthWrapper(lambda: self._get_token()['access_token'])

    def get_current_user(self) -> str:
        return self._username

    def _get_token(self):
        with self._get_token_lock:
            # Refresh access token using refresh token
            if datetime.now() > self.access_token_expiry:
                self._refresh_access_token()

        return self._token_info

    def _refresh_access_token(self):
        try:
            cognito = boto3.client('cognito-idp', region_name=self.region)
            resp = cognito.initiate_auth(
                ClientId=self.client_id,
                AuthFlow='REFRESH_TOKEN_AUTH',
                AuthParameters={
                    'REFRESH_TOKEN': self._token_info['refresh_token']
                }
            )
            logger.debug('Successfully refreshed token')
        except ClientError as err:
            logger.warning(err)
            self._clear_token_info()
            raise RuntimeError('Failed to refresh token, please reauthenticate')

        auth_result = resp['AuthenticationResult']
        self._token_info['access_token'] = auth_result['AccessToken']
        self._token_info['id_token'] = auth_result['IdToken']
        self._save_token_info()
        self._update_token_metadata()

    def _update_token_metadata(self):
        decoded_access_token = jwt.decode(self._token_info['access_token'],
                                          options={"verify_signature": False})
        self.access_token_expiry = datetime.fromtimestamp(decoded_access_token['exp'])
        self._username = decoded_access_token['username']

    def _load_token_info(self) -> Optional[OAuthTokenResponse]:
        if not self._persistence or not TOKEN_PATH.exists():
            return None

        token_info = json.loads(self._persistence.load())
        if 'access_token' not in token_info:
            return None

        return token_info

    def _save_token_info(self):
        if not self._persistence:
            return

        self._persistence.save(json.dumps(self._token_info))

    def _clear_token_info(self):
        if not self._persistence:
            return

        Path(self._persistence.get_location()).unlink(missing_ok=True)
        self._token_info = None



from typing import List

from cirro.api.models.project import Project
from cirro.cli.interactive.common_args import ask_project
from cirro.cli.models import ListArguments


def gather_list_arguments(input_params: ListArguments, projects: List[Project]):

    # Get the project name
    project_name = ask_project(projects, input_params.get('project'))

    # Map the name back to a unique ID
    id_map = {
        project.name: project.id
        for project in projects
    }

    # Save the ID
    input_params['project'] = id_map[project_name]

    return input_params



import questionary
from questionary import prompt


class InputError(Exception):
    """Errors detected from user input"""
    pass


def prompt_wrapper(questions):
    answers = prompt(questions)
    # Prompt catches KeyboardInterrupt and sends back an empty dictionary
    # We want to catch this exception
    if len(answers) == 0:
        raise KeyboardInterrupt()
    return answers


def type_validator(t, v):
    """Return a boolean indicating whether `v` can be cast to `t(v)` without raising a ValueError."""
    try:
        t(v)
        return True
    except ValueError:
        return False


def ask(fname, msg, validate_type=None, output_f=None, **kwargs) -> str:
    """Wrap questionary functions to catch escapes and exit gracefully."""

    # Get the questionary function
    questionary_f = questionary.__dict__.get(fname)

    # Make sure that the function exists
    assert questionary_f is not None, f"No such questionary function: {fname}"

    if fname == "select":
        kwargs["use_shortcuts"] = True

    if validate_type is not None:
        kwargs["validate"] = lambda v: type_validator(validate_type, v)

    # The default value must be a string
    if kwargs.get("default") is not None:
        kwargs["default"] = str(kwargs["default"])

    if kwargs.get("required"):
        del kwargs["required"]
        kwargs["validate"] = lambda val: len(val.strip()) > 0 or "This field is required"

    # Add a spacer line before asking the question
    print("")

    # Get the response
    resp = questionary_f(msg, **kwargs).ask()

    # If the user escaped the question
    if resp is None:
        raise KeyboardInterrupt()

    # If an output transformation function was defined
    if output_f is not None:

        # Call the function
        resp = output_f(resp)

    # Otherwise
    return resp


def ask_yes_no(msg):
    return ask("select", msg, choices=["Yes", "No"]) == "Yes"


def get_id_from_name(items, name_or_id: str) -> str:
    return get_item_from_name_or_id(items, name_or_id).id


def get_item_from_name_or_id(items, name_or_id: str):
    matched = next((p for p in items if p.id == name_or_id), None)
    if matched:
        return matched
    return next(p for p in items if p.name == name_or_id)



from cirro.api.services.common import CommonService
from cirro.api.services.dataset import DatasetService
from cirro.api.services.file import FileService
from cirro.api.services.process import ProcessService
from cirro.api.services.project import ProjectService

__all__ = [
    'DatasetService',
    'ProjectService',
    'ProcessService',
    'CommonService',
    'FileService'
]



import os
import random
import time
from pathlib import Path, PurePath
from typing import List, Union

from boto3.exceptions import S3UploadFailedError
from botocore.exceptions import ConnectionError

from cirro.api.clients import S3Client
from cirro.api.models.file import DirectoryStatistics, File

if os.name == 'nt':
    import win32api
    import win32con

DEFAULT_TRANSFER_SPEED = 160


def filter_files_by_pattern(files: Union[List[File], List[str]], pattern: str) -> List[File]:
    """
    Filters a list of files by a glob pattern
    """
    def matches_glob(file: Union[File, str]):
        return PurePath(file if isinstance(file, str) else file.relative_path).match(pattern)

    return [
        file for file in files
        if matches_glob(file)
    ]


def _is_hidden_file(file_path: Path):
    # Remove hidden files from listing, desktop.ini .DS_Store, etc.
    if os.name == 'nt':
        attributes = win32api.GetFileAttributes(str(file_path))
        return attributes & (win32con.FILE_ATTRIBUTE_HIDDEN | win32con.FILE_ATTRIBUTE_SYSTEM)
    else:
        return file_path.name.startswith('.')


def get_files_in_directory(directory) -> List[str]:
    path = Path(directory)
    path_posix = str(path.as_posix())

    paths = []

    for file_path in path.rglob("*"):
        if file_path.is_dir():
            continue

        if _is_hidden_file(file_path):
            continue

        str_file_path = str(file_path.as_posix())
        str_file_path = str_file_path.replace(f'{path_posix}/', "")
        paths.append(str_file_path)

    return paths


def get_directory_stats(directory) -> DirectoryStatistics:
    sizes = [f.stat().st_size for f in Path(directory).glob('**/*') if f.is_file()]
    total_size = sum(sizes) / float(1 << 30)
    return {
        'sizeFriendly': f'{total_size:,.3f} GB',
        'size': total_size,
        'numberOfFiles': len(sizes)
    }


def upload_directory(directory: str, files: List[str], s3_client: S3Client, bucket: str, prefix: str, max_retries=10):
    for file in files:
        key = f'{prefix}/{file}'
        local_path = Path(directory, file)
        success = False

        # Retry up to max_retries times
        for retry in range(max_retries):

            # Try the upload
            try:
                s3_client.upload_file(
                    local_path=local_path,
                    bucket=bucket,
                    key=key
                )

                success = True

            # Catch the upload error
            except (S3UploadFailedError, ConnectionError) as e:
                delay = random.uniform(0, 60) + retry * 60
                # Report the error
                print(f"Encountered error:\n{str(e)}\n"
                      f"Retrying in {delay:.0f} seconds ({max_retries - (retry + 1)} attempts remaining)")
                time.sleep(delay)

            if success:
                break


def download_directory(directory: str, files: List[str], s3_client: S3Client, bucket: str, prefix: str):
    for file in files:
        key = f'{prefix}/{file}'
        local_path = Path(directory, file)
        local_path.parent.mkdir(parents=True, exist_ok=True)

        s3_client.download_file(local_path=local_path,
                                bucket=bucket,
                                key=key)


def estimate_token_lifetime(data_size_gb: float, speed_mbps: float = DEFAULT_TRANSFER_SPEED) -> int:
    """
    :param data_size_gb: Gigabytes
    :param speed_mbps: Megabits per second
    """
    transfer_time_seconds = (data_size_gb * 8 * 1000) / speed_mbps
    transfer_time_hours = transfer_time_seconds / 60 / 60
    return max(round(transfer_time_hours), 1)



from pathlib import Path
from typing import Union

import github
from github import Github
from github.Branch import Branch
from github.GitRelease import GitRelease
from github.Repository import Repository

from cirro.api.models.workflow_models import OptimizedOutput, Column, WorkflowRepository
from cirro.cli.interactive.utils import ask, ask_yes_no


def get_output_resources_path():
    """Get the location of the folder within the cirro resources repository to write to."""

    # Get the base directory of the repository
    repo_folder = _get_repo_folder()

    # Build the subdirectory for the process
    subdirectory = ask(
        "text",
        "What subdirectory within the process/ folder which should be used to save the outputs?"
        " (e.g. hutch/fastqc/1.0)",
        required=True
    )

    resources_folder = Path(repo_folder, "process", subdirectory)

    # Check that the resources folder already exists and prompt until confirmed
    while resources_folder.exists():

        if ask(
                "select",
                "That folder already exists, would you like to overwrite any existing files or select a new folder?",
                choices=[f"Use {resources_folder}", "Select another"]
        ) == "Select another":

            # Build the subdirectory for the process
            subdirectory = ask(
                "text",
                "What subdirectory within the process/ folder which should be used to save the outputs?"
                " (e.g. hutch/fastqc/1.0)",
                required=True
            )
            resources_folder = Path(repo_folder, "process", subdirectory)

        else:
            break

    return resources_folder, subdirectory


def _get_repo_folder():
    """Get the base location of the cirro resources repository."""

    repo_folder = ask(
        "path",
        "What folder contains a local copy of the Cirro resources repository?",
        default=Path.cwd(),
        only_directories=True
    )
    repo_folder = Path(repo_folder)

    # If the path does not exist
    if not repo_folder.exists():

        # Ask if it should be created
        resp = ask(
            "select",
            f"The path does not exist: {repo_folder}\nWould you like to create it, or pick another folder?",
            choices=[
                "Create the folder",
                "Select another"
            ]
        )

        if resp == "Create the folder":
            repo_folder.mkdir(parents=True)
            return repo_folder

        else:
            return _get_repo_folder()

    else:
        return repo_folder


def get_child_processes(processes):
    # Get a list of the processes which are available
    process_choices = [
        f"{process['name']}\n     {process['desc']}\n     {process['id']}"
        for process in processes
    ]
    process_choices.sort()

    return [
        p.split("\n")[-1].strip(" ")
        for p in ask(
            "checkbox",
            "Select any processes which can be run on the outputs of this workflow",
            choices=process_choices
        )
    ]


def get_outputs():
    outputs = []

    if ask(
            "confirm",
            "Does this workflow produce output files which should be indexed for visualization?"
    ):
        outputs.append(_add_single_output())

        while ask("confirm", "Would you like to add another?"):
            outputs.append(_add_single_output())

    return outputs


def _add_single_output() -> OptimizedOutput:
    # Get the path of the output file(s) relative to the output directory
    source = ask(
        "text",
        "\n  ".join([
            "What is the location of the output files(s) within the workflow output directory?",
            "Multiple files with the same format can be included using wildcard (*) characters.\n"
        ])
    )
    # TODO: automatically get details based on sample file

    # Get the value used to separate columns
    sep = ask(
        "text",
        "\n  ".join([
            "What is the character used to separate columns?",
            "e.g. ',' for CSV, '\\t' for TSV:"
        ])
    )

    name = ask("text", "Short name for output file(s)")
    desc = ask("text", "Longer description for output file(s)")
    url = ask("text", "Optional website documenting file contents")

    # Build the list of columns
    columns = []

    print("")

    while len(columns) == 0 or ask("confirm", "Are there additional columns to add?"):
        columns.append(Column(
            header=ask("text", "Column header (value in the first line of the file)"),
            display_name=ask("text", "Column name (to be displayed to the user)"),
            description=ask("text", "Column description (to be displayed to the user)")
        ))

    if ask(
            "confirm",
            "\n     ".join([
                               "File configuration:",
                               f"Path: {source}",
                               f"Sep: {sep}",
                               f"Name: {name}",
                               f"Description: {desc}",
                               f"Reference URL: {url}",
                               "Columns:"
                           ] + [
                               f"    {col.header} - {col.display_name} - {col.description}"
                               for col in columns
                           ] + [
                               "Is all of the information above correct?"
                           ])
    ):
        return OptimizedOutput(
            source_pattern=source,
            seperator=sep,
            name=name,
            description=desc,
            documentation_url=url,
            columns=columns
        )


def get_preprocess_script():
    """Ask if the user wants to add a preprocessing script."""

    if ask_yes_no("Would you like to use a preprocessing script?"):

        script_path = ask(
            "path",
            "What script should be used?",
            default=Path.cwd()
        )
        script_path = Path(script_path)

        while not script_path.exists():

            script_path = ask(
                "path",
                f"Path cannot be found ({script_path}) - please select another",
                default=script_path
            )

        return script_path


def get_repository() -> WorkflowRepository:
    github_connection = Github()

    name = ask(
        "text",
        "What name should be displayed for this workflow?",
        required=True
    )

    # Get the organization
    org = ask(
        'text',
        'Which GitHub organization is the workflow located within?',
        default='nf-core',
        required=True
    )

    # Get the repository
    repo_name = _prompt_repository(org, github_connection)

    # Get the repository object
    repo = github_connection.get_repo(f"{org}/{repo_name}")

    # Get the version
    version = _prompt_repository_version(repo)
    if isinstance(version, GitRelease):
        documentation_url = version.url
        tarball_url = version.tarball_url
        version_name = version.tag_name
    else:
        documentation_url = f"https://github.com/{org}/{repo_name}/tree/{version.name}"
        tarball_url = ''
        version_name = version.name

    entrypoint = ask(
        "text",
        "What is the primary entrypoint for the workflow in the repository?",
        default="main.nf"
    )
    try:
        repo.get_contents(entrypoint, version_name)
    except github.UnknownObjectException:
        print(f"Warning: {entrypoint} not found in {repo_name}")

    return WorkflowRepository(display_name=name,
                              org=org,
                              repo_name=repo_name,
                              version=version_name,
                              entrypoint=entrypoint,
                              private=repo.private,
                              documentation_url=documentation_url,
                              tarball=tarball_url)


def _prompt_repository(org, github_connection):
    """Prompt the user for a repository contained within an organization."""

    # Get a list of repos in that organization
    user = github_connection.get_user(org)
    repo_count = user.public_repos

    if repo_count < 500:
        repo_list = [
            repo.name for repo in github_connection.get_user(org).get_repos()
            if repo.name != '.github'
        ]

        # then use that to ask the user which repo to look at
        return ask(
            'autocomplete',
            'Which repository contains the workflow of interest? (use TAB to display options)',
            choices=repo_list
        )

    while True:
        repo_name = ask(
            'text',
            'Enter the name of the repository',
            required=True
        )

        try:
            user.get_repo(repo_name)
            return repo_name
        except github.UnknownObjectException:
            print('Repository name not valid')


def _prompt_repository_version(repo: Repository) -> Union[GitRelease, Branch]:
    """Parse the repository and ask the user which tag/version to use."""

    # The version will be specified with either a branch or a release
    version_type = ask(
        'select',
        'Should the workflow version be specified by branch or release tag',
        choices=['branch', 'release']
    )

    # If the user decided to select the version type by release (tag)
    if version_type == 'release':

        # Get the releases which are available
        version_list = repo.get_releases()
        pretty_version_list = [f"{x.tag_name} ({x.title})" for x in version_list]

        version = ask(
            'select',
            f'Which version of {repo.name} do you want to use?',
            choices=pretty_version_list
        )

        version = [x for x in version_list if f"{x.tag_name} ({x.title})" == version][0]
        return version

    else:

        assert version_type == "branch"

        # Select from the branches which are available
        branch = ask(
            'select',
            f'Which branch of {repo.name} do you want to use?',
            choices=[branch.name for branch in repo.get_branches()]
        )

        branch = repo.get_branch(branch)

        return branch


def get_description():
    return ask(
        "text",
        "What is the description of this workflow?"
    )


def get_additional_inputs():
    """Configure any additional inputs."""

    inputs = []
    if ask(
        "confirm",
        "Would you like to add any fixed parameter values?\n  These values will not change based on user input."
    ):
        if (input_param := _add_fixed_param_input()) is not None:
            inputs.append(input_param)

        while ask("confirm", "Would you like to add another?"):
            if (input_param := _add_fixed_param_input()) is not None:
                inputs.append(input_param)

    return inputs


def _add_fixed_param_input():
    """Allow the user to add a single key-value entry to the params."""

    # Allow the user to select some cirro-specific elements
    param_type = ask(
        "select",
        "What type of parameter is this?",
        choices=[
            "Fixed value",
            "Location of data inputs",
            "Location of data outputs",
        ]
    )

    if param_type == "Fixed value":

        # Get a key-value pair
        param_key = ask("text", "Parameter key:")
        param_value = ask("text", "Parameter value:")

    elif param_type == "Location of data inputs":

        param_key = ask("text", "Parameter key:")
        param_value = "$.params.inputs[0].s3|/data/"

    elif param_type == "Location of data outputs":

        param_key = ask("text", "Parameter key:")
        param_value = "$.params.dataset.s3|/data/"

    else:
        raise AssertionError(f"Internal error: parameter type {param_type} is not configured")

    # Let the user confirm their entry
    if ask("confirm", f"Confirm: {param_key} = {param_value}"):
        return param_key, param_value



from typing import Union

from cirro.api.clients.portal import DataPortalClient
from cirro.api.models.dataset import Dataset
from cirro.api.models.process import RunAnalysisCommand
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset
from cirro.sdk.exceptions import DataPortalInputError
from cirro.sdk.file import DataPortalFile, DataPortalFiles
from cirro.sdk.helpers import parse_process_name_or_id
from cirro.sdk.process import DataPortalProcess


class DataPortalDataset(DataPortalAsset):
    """
    Datasets in the Data Portal are collections of files which have
    either been uploaded directly, or which have been output by
    an analysis pipeline or notebook.
    """
    name = None

    def __init__(self, dataset: Dataset, client: DataPortalClient):
        assert dataset.project_id is not None, "Must provide dataset with project_id attribute"
        self.id = dataset.id
        self.name = dataset.name
        self.description = dataset.description
        self.process_id = dataset.process_id
        self.project_id = dataset.project_id
        self.status = dataset.status
        self.source_dataset_ids = dataset.source_dataset_ids
        self.info = dataset.info
        self.params = dataset.params
        self.created_at = dataset.created_at
        self._client = client

    def __str__(self):
        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'id', 'description', 'status']
        ])

    def list_files(self) -> DataPortalFiles:
        """Return the list of files which make up the dataset."""

        return DataPortalFiles(
            [
                DataPortalFile(
                    f,
                    client=self._client
                )
                for f in self._client.dataset.get_dataset_files(
                    project_id=self.project_id,
                    dataset_id=self.id
                )
            ]
        )

    def download_files(self, download_location: str = None) -> None:
        """Download all of the files from the dataset to a local directory."""

        # Alias for internal method
        self.list_files().download(download_location)

    def run_analysis(
            self,
            name: str = None,
            description: str = "",
            process: Union[DataPortalProcess, str] = None,
            params=None,
            notifications_emails=None
    ) -> str:
        """
        Runs an analysis on a dataset, returns the ID of the new dataset.
        The process can be provided as either a DataPortalProcess object,
        or a string which corresponds to the name or ID of the process.
        """
        if name is None:
            raise DataPortalInputError("Must specify 'name' for run_analysis")
        if process is None:
            raise DataPortalInputError("Must specify 'process' for run_analysis")
        if notifications_emails is None:
            notifications_emails = []
        if params is None:
            params = {}

        # If the process is a string, try to parse it as a process name or ID
        process = parse_process_name_or_id(process, self._client)

        return self._client.process.run_analysis(
            RunAnalysisCommand(
                name=name,
                description=description,
                process_id=process.id,
                parent_dataset_id=self.id,
                project_id=self.project_id,
                params=params,
                notifications_emails=notifications_emails
            )
        )


class DataPortalDatasets(DataPortalAssets[DataPortalDataset]):
    """Collection of multiple DataPortalDataset objects."""
    asset_name = "dataset"



from cirro.api.clients.portal import DataPortalClient
from cirro.api.models.process import Executor
from cirro.sdk.process import DataPortalProcess, DataPortalProcesses
from cirro.sdk.project import DataPortalProject, DataPortalProjects
from cirro.sdk.reference_type import DataPortalReferenceType, DataPortalReferenceTypes


class DataPortal:
    """
    Helper functions for exploring the Projects, Datasets, Samples, and Files
    available in the Data Portal.
    """

    def __init__(self, client: DataPortalClient = None):
        """Set up the DataPortal object, establishing an authenticated connection."""

        if client is not None:
            self._client = client

        # Set up default client if not provided
        else:
            self._client = DataPortalClient()

    def list_projects(self) -> DataPortalProjects:
        """List all of the projects available in the Data Portal."""

        return DataPortalProjects(
            [
                DataPortalProject(proj, self._client)
                for proj in self._client.project.list()
            ]
        )

    def get_project_by_name(self, name: str = None) -> DataPortalProject:
        """Return the project with the specified name."""

        return self.list_projects().get_by_name(name)

    def get_project_by_id(self, _id: str = None) -> DataPortalProject:
        """Return the project with the specified id."""

        return self.list_projects().get_by_id(_id)

    def list_processes(self, ingest=False) -> DataPortalProcesses:
        """
        List all of the processes available in the Data Portal.
        By default, only list non-ingest processes (those which can be run on existing datasets).
        To list the processes which can be used to upload datasets, use ingest = True.
        """

        return DataPortalProcesses(
            [
                DataPortalProcess(p, self._client)
                for p in self._client.process.list()
                if (p.executor == Executor.INGEST) == ingest
            ]
        )

    def get_process_by_name(self, name: str, ingest=False) -> DataPortalProcess:
        """Return the process with the specified name."""

        return self.list_processes(ingest=ingest).get_by_name(name)

    def get_process_by_id(self, _id: str, ingest=False) -> DataPortalProcess:
        """Return the process with the specified id."""

        return self.list_processes(ingest=ingest).get_by_id(_id)

    def list_reference_types(self) -> DataPortalReferenceTypes:
        """Return the list of all available reference types."""

        return DataPortalReferenceTypes(
            [
                DataPortalReferenceType(ref)
                for ref in self._client.common.get_references_types()
            ]
        )



from dataclasses import dataclass

from cirro.api.models.exceptions import DataPortalModelException


@dataclass(frozen=True)
class Project:
    id: str
    name: str
    description: str

    @classmethod
    def from_record(cls, record):
        if record is None:
            raise DataPortalModelException("Cannot construct Dataset from null object.")
        return cls(record['id'],
                   record['name'],
                   record['desc'])

    @property
    def relative_url(self):
        return f'projects/{self.id}'



import json
from datetime import timezone, datetime
from typing import Optional, Union

from cirro.api.models.auth import Creds


def parse_json_date(json_date: str) -> Optional[datetime]:
    if not json_date:
        return None
    local_zone = datetime.now(timezone.utc).astimezone().tzinfo
    try:
        parsed = datetime.strptime(json_date, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        # AWS Format
        parsed = datetime.strptime(json_date, "%Y-%m-%dT%H:%M:%S+00:00")

    return parsed\
        .replace(tzinfo=timezone.utc)\
        .astimezone(local_zone)


def safe_load_json(json_str: Optional[str]):
    if json_str is None:
        return {}

    return json.loads(json_str)


def format_date(date: Union[str, datetime]) -> str:
    if isinstance(date, str):
        date = parse_json_date(date)
    return date.strftime('%m/%d/%Y, %H:%M:%S')


def print_credentials(creds: Creds):
    print(f'AWS_ACCESS_KEY_ID: {creds["AccessKeyId"]}')
    print(f'AWS_SECRET_ACCESS_KEY: {creds["SecretAccessKey"]}')
    print(f'AWS_SESSION_TOKEN: {creds["SessionToken"]}')
    if creds['Expiration']:
        print()
        print(f'These credentials expire at {format_date(creds["Expiration"])}')



import json

from github import UnknownObjectException, Github

from cirro.cli.interactive.utils import ask, ask_yes_no
from cirro.helpers.constants import IGENOMES_REFERENCES


def prompt_user_inputs():
    """Prompt for user inputs which should be included in the form."""

    # Make a dict with the properties which should be rendered in the form
    # Each parameter will fall inside a section with its own title
    properties = dict()
    inputs = dict()

    print("The form which is presented to the user is broken into sections, each with its own heading")

    while len(properties) == 0 or ask("confirm", "Would you like to add another section of parameters?"):
        # Get the contents of this section
        section_key, section_contents = _prompt_user_input_section(inputs)

        # Add those contents to the form
        properties[section_key] = section_contents

    return properties, inputs


def _prompt_user_input_section(inputs):
    """Ask the user for all of the parameters which are found in a single section of the input form."""

    section_title = ask("text", "What is the title of this section?")

    # Format the section key based on the title
    section_key = section_title.lower().replace(" ", "_").replace("-", "_")

    # Build a dict of the individual form entries
    properties = dict()

    # Keep track of which inputs are required
    required = []

    for input_key, input_val, input_required in _yield_form_input_single(inputs, section_key):

        properties[input_key] = input_val
        if input_required:
            required.append(input_key)

    # Return the fully formulated section
    section_contents = dict(
        type="object",
        title=section_title,
        required=required,
        properties=properties
    )
    return section_key, section_contents


def _yield_form_input_single(inputs, section_key):
    """Allow the user to add as many form inputs as they require."""

    initial = True

    while initial or ask("confirm", "Would you like to add another parameter to this section?"):
        initial = False

        yield _prompt_form_input_single(inputs, section_key)


def _prompt_form_input_single(inputs, section_key):
    """Prompt the user for a single element in the form."""

    # Prompt for the parameter key
    kw = ask("text", "Parameter key used in workflow:")

    # There should be no collision between keys
    while len(kw) == 0 or kw in inputs:
        kw = ask("text", "Parameter key is not valid or has already been used, please select another:")

    elem = dict()
    elem["title"] = ask("text", "Parameter title (optional)")
    elem["description"] = ask("text", "Parameter description:")
    required = ask_yes_no("Is this parameter required?")

    # Get the type of input
    prompt_type = ask(
        "select",
        "What type of input should the user provide?",
        choices=[
            "Text entry",
            "Choose from list",
            "Boolean",
            "Number",
            "File from the input dataset",
            "File from a project reference",
            "List of iGenomes references",
        ]
    )

    if prompt_type == "Text entry":
        elem["type"] = "string"
        if ask_yes_no("Does this input have a default value?"):
            elem["default"] = ask("text", "Default value: ")

    elif prompt_type == "Choose from list":
        print("Please set up the options which the user can choose between")
        print("Note that you can customize the value displayed to the user")
        elem["type"] = "string"
        elem["enum"] = []
        elem["enumNames"] = []

        while len(elem["enum"]) == 0 or ask_yes_no("Would you like to add another option?"):
            val = ask("text", "Value:")
            name = ask("text", "Display:", default=val)

            elem["enum"].append(val)
            elem["enumNames"].append(name)

    elif prompt_type == "Boolean":
        elem["type"] = "boolean"
        if ask_yes_no("Does this input have a default value?"):
            elem["default"] = ask("select", "Default value:", choices=["True", "False"]) == "True"

    elif prompt_type == "Number":
        if ask_yes_no("Does this input have a default value?"):
            if ask("select", "Integer or float?", choices=["Integer", "Float"]) == "Integer":
                elem["default"] = int(ask("text", "Default value:", validate_type=int))
                elem["type"] = "integer"
            else:
                elem["default"] = float(ask("text", "Default value:", validate_type=float))
                elem["type"] = "number"

    elif prompt_type == "List of iGenomes references":
        elem["type"] = "string"
        elem["enum"] = IGENOMES_REFERENCES.keys()
        elem["enumNames"] = IGENOMES_REFERENCES.values()

    elif prompt_type == "File from the input dataset":

        elem["type"] = "string"
        elem["file"] = f"**/*{ask('text', 'What is the expected file extension?')}"
        elem["pathType"] = "dataset"
        elem["mm"] = dict(matchBase=True)

    elif prompt_type == "File from a project reference":

        elem["type"] = "string"
        elem["pathType"] = "references"

        ref_title = ask("text", "What is the name of the reference type?")
        ref_file = ask("text", "What is the name of the file within the reference?")
        elem["file"] = f"**/{ref_title}/**/{ref_file}"

    else:
        assert f"Internal error: prompt type not configured: {prompt_type}"

    # Map the form entry to the input params
    inputs[kw] = f"{section_key}.{kw}"

    # Return the configuration of the parameter
    return kw, elem, required


def get_nextflow_schema(repo: str, ref: str):
    """If a nextflow_schema.json is present, parse it and prompt the user for any modifications."""
    gh = Github()
    repo = gh.get_repo(repo)
    try:
        schema_file = repo.get_contents('nextflow_schema.json', ref)
        nf_schema = json.loads(schema_file.decoded_content)

        # Make sure that the schema has the expected top-level entries
        for k in ["title", "definitions"]:
            assert k in nf_schema, f"Did not find '{k}' in the nextflow schema as expected"

        return nf_schema
    except UnknownObjectException:
        print("No nextflow_schema.json file found, please specify any required inputs")
        return None


def convert_nf_schema(form_obj, inputs, param_root="$.params.dataset.paramJson"):
    """
    Given a nextflow schema object, parse it and
    run the user through prompts to determine which fields to expose
    """

    # Remove any unwanted keys
    for k in ["$schema", "$id", "fa_icon", "mimetype", "allOf", "schema"]:
        if k in form_obj:
            del form_obj[k]

    # Convert the 'definitions' from the Nextflow schema to 'properties' of the form
    if "definitions" in form_obj:
        form_obj["properties"] = form_obj["definitions"]
        del form_obj["definitions"]

    # The object should have a "type"
    assert "type" in form_obj, f"Expected 'type':\n({json.dumps(form_obj, indent=4)})"

    # If this is an object
    if form_obj["type"] == "object":

        # It must have "properties"
        assert "properties" in form_obj, f"Expected 'properties' when type='object':\n" \
                                         f"({json.dumps(form_obj, indent=4)})"

        # Give the user the option to remove items from that list

        # First ask the user which options should be kept
        option_list = [
            f"{k}\n     {v.get('description')}"
            for k, v in form_obj["properties"].items()
        ]
        to_keep = ask(
            "checkbox",
            "Please select the items to display to the user\n<space> to select/deselect, <enter> when done",
            choices=option_list
        )

        # Remove the description from the list of
        to_keep = [i.split("\n")[0] for i in to_keep]

        # Subset the defs to only keep the selected options
        form_obj["properties"] = {
            k: form_obj["properties"][k]
            for k in to_keep
        }

        # If there is a "required" field at the top level
        if "required" in form_obj:
            # Only keep the required items which were selected
            form_obj["required"] = [i for i in form_obj["required"] if i in to_keep]

        # Now convert each of those objects in a recursive way
        form_obj["properties"] = {
            k: convert_nf_schema(v, param_root=f"{param_root}.{k}", inputs=inputs)
            for k, v in form_obj["properties"].items()
        }

        # For each of those elements
        for k, v in form_obj["properties"].items():

            # Skip any other objects
            if v["type"] == "object":
                continue

            # Any other type needs to be added to the inputs

            # Make sure that this parameter name doesn't collide with anything else
            assert k not in inputs, f"Error, found two elements in the form named {k}"

            # Add it to the inputs
            inputs[k] = f"{param_root}.{k}"

    # Return the object
    return form_obj



from cirro.helpers.form import FormBuilder
from cirro.helpers.preprocess_dataset import PreprocessDataset
from cirro.helpers.workflow_config import WorkflowConfigBuilder

__all__ = [
    'WorkflowConfigBuilder',
    'PreprocessDataset',
    'FormBuilder'
]



from enum import Enum


class Status(Enum):
    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    DELETE = 'DELETE'
    DELETING = 'DELETING'
    FAILED = 'FAILED'
    COMPLETED = 'COMPLETED'
    ARCHIVED = 'ARCHIVED'
    DELETED = 'DELETED'
    SUSPENDED = 'SUSPENDED'



import json
import shutil
from pathlib import Path
from typing import List, Dict

from cirro.api.models.process import ProcessRecord, Executor
from cirro.api.models.workflow_models import OptimizedOutput, WorkflowRepository, ProcessConfig
from cirro.helpers.constants import PROCESSES_PATH_S3


class WorkflowConfigBuilder:
    def __init__(self, repo_prefix: str):
        """
        Initializes workflow config builder
        :param repo_prefix: subdirectory to write to (i.e. hutch/fastqc/1.0
        """

        # All the parameters will be added to a single object
        self.process_config = ProcessConfig(
            dynamo=dict(),
            form=dict(),
            input=dict(),
            output=dict()
        )
        self.repo_prefix = repo_prefix.strip('/')
        self.compute_config = ""
        self.preprocess_py_path = None

        # Set up the boilerplate elements of the dynamo record
        self._add_dynamo_boilerplate()

    def save_local(self, output_folder: Path):
        """
        Write out the workflow configuration as a collection of files.
        """
        output_folder.mkdir(parents=True, exist_ok=True)

        # Reorder dynamo config based on model order
        ordered_record = {}
        for key in ProcessRecord.__annotations__.keys():
            value = self.process_config["dynamo"].get(key)  # type: ignore
            if value is not None:
                ordered_record[key] = value
        self.process_config["dynamo"] = ordered_record

        # Save each of the items in the process configuration
        for config_name, config_value in self.process_config.items():
            output_fp = Path(output_folder, f"process-{config_name}.json")
            print(f"Writing out to {output_fp}")

            with output_fp.open('w') as handle:
                json.dump(config_value, handle, indent=4)

        # Write the compute configuration
        compute_path = Path(output_folder, "process-compute.config")
        print(f"Boilerplate compute configuration has been written to {compute_path}"
              f" -- please modify that file as necessary.")

        with compute_path.open("w") as handle:
            handle.write(self.compute_config)

        # Copy preprocess file
        if self.preprocess_py_path is not None:
            preprocess_dest = Path(output_folder, self.preprocess_py_path.name)
            print(f"Copying preprocess from {self.preprocess_py_path} to {preprocess_dest}")
            shutil.copyfile(
                self.preprocess_py_path,
                preprocess_dest
            )
        print(f"Done writing all process configuration items to {output_folder}")

    def with_description(self, description: str):
        self.process_config["dynamo"]["desc"] = description
        return self

    def with_repository(self, repo: WorkflowRepository):
        """
        Configure the workflow repository.
        """

        # Get the name of the process
        self.process_config["dynamo"]["name"] = repo.display_name

        # Set up the process name based on the workflow/version
        process_id = f"process-{repo.org}-{repo.repo_name}-{repo.version}".lower()
        self.process_config["dynamo"]["id"] = process_id

        repository_code = f"GITHUB{'PRIVATE' if repo.private else 'PUBLIC'}"
        # Set up the 'code' block of the dynamo record
        self.process_config["dynamo"]["code"] = dict(
            repository=repository_code,
            script=repo.entrypoint,
            uri=f"{repo.org}/{repo.repo_name}",
            version=repo.version
        )
        self.process_config["dynamo"]["documentationUrl"] = repo.documentation_url

        return self

    def with_child_processes(self, child_processes: List[str]):
        """
        Specify what child processes are allowed to run from the output of this dataset
        """
        self.process_config["dynamo"]["childProcessIds"] = child_processes

        return self

    def with_preprocess(self, preprocess_py_path: Path):
        """
        Add a preprocess file
        """
        # Add it to the dynamo record
        self.process_config["dynamo"]["preProcessScript"] = \
            f"{PROCESSES_PATH_S3}/{self.repo_prefix}/{preprocess_py_path.name}"
        self.preprocess_py_path = preprocess_py_path

        return self

    def _add_dynamo_boilerplate(self):
        """
        Add the elements of the dynamo record which do not vary by user entry.
        """

        self.process_config["dynamo"]["executor"] = Executor.NEXTFLOW.value
        self.process_config["dynamo"]["paramDefaults"] = []
        self.process_config["dynamo"]["fileJson"] = ""
        self.process_config["dynamo"]["componentJson"] = ""
        self.process_config["dynamo"]["infoJson"] = ""
        self.process_config["dynamo"]["paramMapJson"] = \
            f"{PROCESSES_PATH_S3}/{self.repo_prefix}/process-input.json"
        self.process_config["dynamo"]["formJson"] = \
            f"{PROCESSES_PATH_S3}/{self.repo_prefix}/process-form.json"
        self.process_config["dynamo"]["webOptimizationJson"] = \
            f"{PROCESSES_PATH_S3}/{self.repo_prefix}/process-output.json"

    def with_compute(self, max_retries=2):
        """
        Configure the compute configuration.
        The compute configuration is boilerplate at this point
        """

        self.process_config["dynamo"]["computeDefaults"] = [
            {
                "executor": "NEXTFLOW",
                "json": f"{PROCESSES_PATH_S3}/{self.repo_prefix}/process-compute.config",
                "name": "Default"
            }
        ]

        self.compute_config = f"""profiles {{
    standard {{
        process {{
            executor = 'awsbatch'
            errorStrategy = 'retry'
            maxRetries = {max_retries}
        }}
    }}
}}
"""
        return self

    def with_form_inputs(self, form):
        """
        Configure the form. UI customizations not supported yet.
        """
        # TODO This could be improved to take a list of form objects
        #  instead of a json schema

        self.process_config["form"] = dict(
            ui=dict(),
            # In addition to modifying the schema contents, we will
            # also ask the user if they want to include or exclude each option
            form=form
        )
        return self

    def with_inputs(self, inputs: Dict[str, str]):
        """
        Configure multiple inputs
        """
        self.process_config["input"] = {
            **self.process_config["input"],
            **inputs
        }
        return self

    def with_input(self, input_name: str, input_value: str):
        """
        Configure a single input parameter mapping
        """
        self.process_config["input"][input_name] = input_value
        return self

    def with_output(self, output: OptimizedOutput):
        """
        Configure a single output file.
        """
        self._init_outputs()
        self.process_config["output"]["commands"].append(
            dict(
                command="hot.Dsv",
                params=dict(
                    url=output.documentation_url,
                    source=output.source_pattern,
                    sep=output.seperator,
                    header=True,
                    name=output.name,
                    desc=output.description,
                    cols=[
                        {
                            'col': col.header,
                            'name': col.display_name,
                            'desc': col.description
                        }
                        for col in output.columns
                    ]
                )
            )
        )
        return self

    def with_common_outputs(self):
        self._init_outputs()
        commands = self.process_config["output"]["commands"]

        # Add hot.Manifest if hot.Dsv is present
        if any(entry['command'] == 'hot.Dsv' for entry in commands):
            # Add the command to index datasources that are present
            commands.append(
                {
                    "command": "hot.Manifest",
                    "params": {}
                }
            )

        # Add the command to index all the files which were outputted
        commands.append(
            {
                "command": "save.ManifestJson",
                "params": {
                    "files": [
                        {
                            "glob": "$data_directory/**/*.*"
                        }
                    ],
                    "tables": [],
                    "jsons": [],
                    "lists": [],
                    "tensors": [],
                    "version": "2"
                }
            }
        )
        return self

    def _init_outputs(self):
        """Checks if we need to initialize the blank output"""
        if not self.process_config["output"]:
            self.process_config["output"] = dict(commands=[])
            self.process_config["dynamo"]["webOptimizationJson"] = \
                f"{PROCESSES_PATH_S3}/{self.repo_prefix}/process-output.json"



from dataclasses import dataclass
from pathlib import PurePath
from typing import Literal, TypedDict, Optional, List

from cirro.api.models.api import ApiQuery

AccessType = Literal['PROJECT', 'CHART', 'DATASET', 'RESOURCES']
FileOperation = Literal['UPLOAD', 'DOWNLOAD']
_GET_FILE_ACCESS_TOKEN_QUERY = '''
  query GetFileAccessToken($input: GetFileAccessTokenInput!) {
    getFileAccessToken(input: $input) {
      AccessKeyId
      Expiration
      SecretAccessKey
      SessionToken
    }
  }
'''


def get_project_bucket(project_id):
    return f'project-{project_id}'


class S3AuthorizerInput(TypedDict):
    accessType: AccessType
    operation: FileOperation
    projectId: Optional[str]
    datasetId: Optional[str]
    tokenLifetimeHours: Optional[int]


class DirectoryStatistics(TypedDict):
    size: float
    sizeFriendly: str
    numberOfFiles: int


class CheckDataTypesInput(TypedDict):
    fileNames: List[str]
    processId: str
    sampleSheet: str


class FileAccessContext:
    """
    Context holder for accessing various files in Cirro and abstracting out their location
    Prefer to use the class methods to instantiate
    """
    def __init__(self,
                 access_input: S3AuthorizerInput,
                 bucket: str,
                 path: str = None):
        self._access_input = access_input
        self.bucket = bucket
        self._path = path

    @classmethod
    def download_dataset(cls, dataset_id: str, project_id: str):
        return cls(
            {
                'accessType': 'PROJECT', 'operation': 'DOWNLOAD',
                'datasetId': dataset_id, 'projectId': project_id,
                'tokenLifetimeHours': None
            },
            get_project_bucket(project_id),
            f'datasets/{dataset_id}'
        )

    @classmethod
    def upload_dataset(cls, dataset_id: str, project_id: str, token_lifetime_override: int = None):
        return cls(
            {
                'accessType': 'DATASET', 'operation': 'UPLOAD',
                'datasetId': dataset_id, 'projectId': project_id,
                'tokenLifetimeHours': token_lifetime_override
            },
            get_project_bucket(project_id),
            f'datasets/{dataset_id}/data'
        )

    @classmethod
    def download_project_resources(cls, project_id: str):
        return cls(
            {
                'accessType': 'PROJECT', 'operation': 'DOWNLOAD',
                'projectId': project_id, 'datasetId': None,
                'tokenLifetimeHours': None
            },
            get_project_bucket(project_id),
            'resources'
        )

    @classmethod
    def upload_project_resources(cls, project_id: str):
        return cls(
            {
                'accessType': 'PROJECT', 'operation': 'UPLOAD',
                'projectId': project_id, 'datasetId': None,
                'tokenLifetimeHours': None
            },
            get_project_bucket(project_id),
            'resources/data'
        )

    @classmethod
    def resources(cls, resources_bucket):
        return cls(
            {
                'accessType': 'RESOURCES', 'operation': 'DOWNLOAD',
                'projectId': None, 'datasetId': None,
                'tokenLifetimeHours': None
            },
            resources_bucket,
            ''
        )

    @property
    def get_token_query(self) -> ApiQuery:
        """ Used to fetch access token """
        return ApiQuery(
            query=_GET_FILE_ACCESS_TOKEN_QUERY,
            variables={'input': self._access_input}
        )

    @property
    def path_prefix(self):
        return self._path

    @property
    def domain(self):
        return f's3://{self.bucket}/{self.path_prefix}'


@dataclass(frozen=True)
class File:
    relative_path: str
    size: int
    access_context: FileAccessContext

    @classmethod
    def of(cls, file: 'File'):
        return cls(file.relative_path, file.size, file.access_context)

    @property
    def absolute_path(self):
        return f'{self.access_context.domain}/{self.relative_path.strip("/")}'

    @property
    def name(self):
        return PurePath(self.absolute_path).name

    def __repr__(self):
        return f'{self.__class__.__name__}(path={self.relative_path})'



from typing import Union

from gql.transport.exceptions import TransportQueryError

from cirro.api.clients.portal import DataPortalClient
from cirro.api.models.exceptions import DataPortalModelException
from cirro.api.models.process import Process
from cirro.sdk.exceptions import DataPortalInputError
from cirro.sdk.process import DataPortalProcess


def parse_process_name_or_id(process: Union[DataPortalProcess, str], client: DataPortalClient):
    """
    If the process is a string, try to parse it as a process name or ID.
    """

    # If the process object is already a DataPortalProcess object
    if isinstance(process, DataPortalProcess):
        return process

    # Otherwise, it should be a string
    if not isinstance(process, str):
        raise DataPortalInputError(f"Process name or ID should be a string: '{process}'")

    # Try to get the process by ID
    try:
        process = client.process.get_process(process)
        if isinstance(process, Process):
            return DataPortalProcess(process, client)

    # Catch the error if no dataset is found
    except (TransportQueryError, DataPortalModelException):
        pass

    # If that didn't work, try to parse it as a name
    try:
        process = client.process.find_by_name(process) or process
        if isinstance(process, Process):
            return DataPortalProcess(process, client)

    # Catch the error if no dataset is found
    except (TransportQueryError, DataPortalModelException):
        pass

    # If that didn't work, raise an error indicating that the process couldn't be parsed
    raise DataPortalInputError(f"Could not parse process name or id: '{process}'")



from typing import List

from cirro.api.models.project import Project
from cirro.cli.interactive.utils import ask


def ask_project(projects: List[Project], input_value: str) -> str:
    project_names = sorted([project.name for project in projects])
    if len(project_names) <= 10:
        return ask(
            'select',
            'What project is this dataset associated with?',
            choices=project_names,
            default=input_value if input_value in project_names else None
        )
    else:
        return ask(
            'autocomplete',
            'What project is this dataset associated with? (use TAB to display options)',
            choices=project_names,
            default=input_value if input_value in project_names else ''
        )



from dataclasses import dataclass
from pathlib import PurePath
from typing import Dict, Any, List, Optional

from cirro.api.models.exceptions import DataPortalModelException
from cirro.api.models.file import File
from cirro.utils import safe_load_json


@dataclass(frozen=True)
class ReferenceType:
    name: str
    description: str
    directory: str
    validation: Dict[str, Any]

    @classmethod
    def from_record(cls, record):
        if record is None:
            raise DataPortalModelException("Cannot construct Dataset from null object.")
        validation = safe_load_json(record['validation'])
        return cls(record['name'], record['description'], record['directory'], validation)

    @property
    def id(self):
        return self.directory

    def __repr__(self):
        fields = f'name={self.name}'
        fields += f', description={self.description}'
        return f'{self.__class__.__name__}({fields})'


class Reference(File):
    @property
    def name(self):
        return PurePath(self.relative_path).parent.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.name == other.name

    def __hash__(self):
        return hash(self.name)


class References(List[Reference]):
    def find_by_name(self, name: str) -> Optional[Reference]:
        return next((ref for ref in self if ref.name == name), None)



from dataclasses import dataclass
from datetime import datetime
from typing import TypedDict, List, Any, Dict

from cirro.api.models.exceptions import DataPortalModelException
from cirro.api.models.status import Status
from cirro.utils import parse_json_date, safe_load_json


class DatasetCreateResponse(TypedDict):
    datasetId: str
    dataPath: str


@dataclass
class CreateIngestDatasetInput:
    name: str
    description: str
    process_id: str
    project_id: str
    files: List[str]

    def to_json(self):
        return {
            'name': self.name,
            'description': self.description,
            'processId': self.process_id,
            'projectId': self.project_id,
            'files': self.files
        }


@dataclass(frozen=True)
class Dataset:
    id: str
    name: str
    description: str
    process_id: str
    project_id: str
    status: Status
    source_dataset_ids: List[str]
    info: Dict[str, Any]
    params: Dict[str, Any]
    created_at: datetime

    @classmethod
    def from_record(cls, record):
        if record is None:
            raise DataPortalModelException("Cannot construct Dataset from null object.")
        return cls(
            record.get('id'),
            record.get('name'),
            record.get('desc'),
            record.get('process'),
            record.get('project'),
            Status[record.get('status')],
            record.get('sourceDatasets'),
            safe_load_json(record.get('infoJson')),
            safe_load_json(record.get('paramJson')),
            parse_json_date(record.get('createdAt'))
        )

    @property
    def relative_url(self):
        return f'projects/{self.project_id}/dataset/{self.id}'
